unit UnitJScript;

interface


uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, AscrLib, ActiveX, ComObj, StdCtrls, ObjComAuto, unitPaste;


type TJScript = class(TComponent, IActiveScriptSite)
   private
        fScript: IActiveScript;
        fParse: IActiveScriptParse;
        terminated : boolean;
        fPaste : TPaste;
        fLastErr : string;
        fLastLine : Cardinal;
        fLastPos : LongInt;


        // IActiveScriptSite interface stuff
        function  GetLCID(out plcid: LongWord): HResult; stdcall;
        function  GetItemInfo(pstrName: PWideChar;
                              dwReturnMask: LongWord;
                              out ppiunkItem: IUnknown;
                              out ppti: IUnknown): HResult; stdcall;
        function  GetDocVersionString(out pbstrVersion: WideString): HResult; stdcall;
        function  OnScriptTerminate(var pvarResult: OleVariant;
                                    var pexcepinfo: EXCEPINFO): HResult; stdcall;
        function  OnStateChange(ssScriptState: tagSCRIPTSTATE): HResult; stdcall;
        function  OnScriptError(const pscripterror: IActiveScriptError): HResult; stdcall;
        function  OnEnterScript: HResult; stdcall;
        function  OnLeaveScript: HResult; stdcall;

        //
        procedure RunScript(script : string);
        function getLastErr : string;

    public
        constructor Create();

        class function isJavaScript(str : string) : boolean;
        class function getStartMacro : string;
        class function getEndMacro : string;
        class function getStartNoMacro : string;
        class function getEndNoMacro : string;

        function executeJS(paste : TPaste; javascript : string) : boolean;

        property LastErr : string read getLastErr;
        property LastLine : cardinal read fLastLine;
        property LastPos : integer read fLastPos;
end;

implementation

uses Clipbrd, UnitClipQueue, System.StrUtils, UnitToken, UnitClipboardGrabber;

const
    CLSID_JScript: TGUID = '{f414c260-6ac0-11cf-b6d1-00aa00bbbb58}';
    SCRIPT_E_REPORTED = HRESULT($80020101);
    ArsClipObject = '_ArsClip';


{$M+}
type TScriptObject = class(TComponent)
    private
        fPaste : TPaste;
        findex : integer;
        function getClipboard : string;
        procedure setClipboard(value : string);
        function getPopupClip : string;
        procedure setPopupIndex(value : Integer);
        procedure RunMacro(macro : string);
    protected
    public
    published
        // these are the properties that are exposed to JavaScript
        property Macro : string write RunMacro;
        property Clipboard : string read getClipboard write setClipboard;
        property PopupClip : string read getPopupClip;
        property PopupIndex : integer write setPopupIndex;
end;
{$M-}
function TScriptObject.getClipboard : string;
begin
    result := TClipboardGrabber.asText;
end;
procedure TScriptObject.setClipboard(value : string);
var
   wc : WideChar;
begin
    clipbrd.Clipboard.SetTextBuf(@value[1]);
end;
procedure TScriptObject.RunMacro(macro : string);
begin
    paste.SendMacro(macro);
end;
procedure TScriptObject.setPopupIndex(value : Integer);
begin
    fIndex := value;
end;
function TScriptObject.getPopupClip : string;
begin
    result := '';
    try
        result := ClipQueue.GetItemText(findex);
    except
    end;
end;
var ScriptObject : TScriptObject;

constructor TJScript.Create();
begin
    inherited create(Application);

    ScriptObject := TScriptObject.Create(Application);

    CoInitializeEx(nil, COINIT_APARTMENTTHREADED);
    OleCheck(CoCreateInstance(CLSID_JScript, nil, CLSCTX_INPROC_SERVER, IID_IActiveScript, FScript));
    OleCheck(FScript.SetScriptSite(Self as IActiveScriptSite));
    OleCheck(FScript.AddNamedItem(ArsClipObject, SCRIPTITEM_ISVISIBLE or SCRIPTITEM_ISSOURCE));

    if Supports(FScript, IActiveScriptParse, FParse) then OleCheck(FParse.InitNew);
end;

const
JS_NL = #13#10;
NL = JS_NL;

JS_MAIN_FUNCTION = 'function main(clipboardStr)';
JS_MAIN_HEADER =
    JS_MAIN_FUNCTION + ' {' + JS_NL +
    '// see [View > JavaScript Reference] for special functions for ArsClip' + JS_NL ;
JS_MAIN_FOOTER =
    '}';
JS_SCRIPTNAME = 'main.js';
JS_SCRIPT_FOOTER =
    NL +
    'function manuallyExecuteMacro(macro) {_ArsClip.Macro = macro;} ' + NL +
    'function saveToPopupClips(clip) {_ArsClip.Macro = ''[newclip=\"''+clip.replace(''"'',''""'')+''\"]'';}' +NL +
    'function getCurrentClipboard() { return _ArsClip.Clipboard;}' +NL+
    'function setClipboard(clip) { _ArsClip.Clipboard = clip; }'+NL+
    'function getPopupClip(index) { _ArsClip.PopupIndex = index; return _ArsClip.PopupClip;}'+NL+
    'main(_ArsClip.Clipboard);';
const
    JS_START_MACRO_VAR = 'start_ac_macro:';
    JS_END_MACRO_VAR = 'end_ac_macro:';
    JS_START_MACRO = JS_START_MACRO_VAR+#13#10;
    JS_END_MACRO = JS_END_MACRO_VAR+#13#10;


class function TJScript.isJavaScript(str : string) : boolean;
var i : integer;
begin
    i := ansiPos(JS_MAIN_FUNCTION, str);
    result := (i > 0);
end;

class function TJScript.getStartMacro : string;
begin
    result := JS_MAIN_HEADER + JS_START_MACRO;
end;
class function TJScript.getEndMacro : string;
begin
    result := JS_END_MACRO + JS_MAIN_FOOTER;
end;
class function TJScript.getStartNoMacro : string;
begin
    result := JS_MAIN_HEADER;
end;
class function TJScript.getEndNoMacro : string;
begin
    result := JS_MAIN_FOOTER;
end;

function TJScript.getLastErr : string;
begin
    result := fLastErr;
end;



function TJScript.executeJS(paste : TPaste; javascript : string) : boolean;
    function encodeACMacros(javascript : string) : string;
    var
        s : string;
        macro : string;
        function createJSLitteralString(const str : string) : string;
        begin
            result := str;
            if length(result) = 0 then begin
                result := '""';
                EXIT;
            end;

            if result[length(result)] = #0 then begin
                delete(result,length(result),1);
            end;
            result := ReplaceText(result, '\', '\\');
            result := ReplaceText(result, '''', '\''');
            result := ReplaceText(result, '"', '\"');
            result := ReplaceText(result, #13, '\r');
            result := ReplaceText(result, #10, '\n');
            result := '"' + result + '"';
        end;
        function MacroToJS(const macro : string) : string;
        var
            s : string;
        begin
            s := macro;
            if macro[length(macro)] = #10 then begin
                delete(s,length(s)-1, 2);
            end;
            s := createJSLitteralString(s);
            result := 'manuallyExecuteMacro('+ s +');' + NL;
        end;
    begin
        while (javascript <> '') do begin
            result := result + TokenString(javascript,JS_START_MACRO_VAR,false);
            if (javascript <> '') then begin
                TokenString(javascript, JS_NL,false);
                macro := TokenString(javascript,JS_END_MACRO_VAR,false);
                TokenString(javascript, JS_NL,false);
                result := result + macroToJS(macro);
            end;
        end;
    end;
begin
    terminated := false;
    fLastErr := '';
    ScriptObject.fPaste := paste;
    javascript := encodeACMacros(javascript);
    javascript := javascript + JS_SCRIPT_FOOTER;

    RunScript(javascript);
    while not terminated do begin
        Sleep(50);
    end;
end;
procedure TJScript.RunScript(script : string);
var
    info: EXCEPINFO;
begin
    try
        OleCheck(FParse.ParseScriptText(PWideChar(script), nil, nil, nil, 0, 0,
            SCRIPTITEM_ISVISIBLE or SCRIPTITEM_ISPERSISTENT, nil, Info));
        OleCheck(FScript.SetScriptState(SCRIPTSTATE_CONNECTED));
        OleCheck(FScript.SetScriptState(SCRIPTSTATE_DISCONNECTED));
    except on e : Exception do
        begin
            terminated := true;
            if (fLastErr = '') then begin
                flastErr := e.Message;
            end;
        end;
    end;
end;





// IActiveScriptSite interface stuff
function TJScript.GetDocVersionString(out pbstrVersion: WideString): HResult;
begin
    pbstrVersion := '1.0';
    Result := S_OK;
end;
function TJScript.GetItemInfo(pstrName: PWideChar; dwReturnMask: LongWord; out ppiunkItem, ppti: IInterface): HResult;
begin
    result := S_FALSE;

    if SameText(ArsClipObject, pstrName) then begin
        if dwReturnMask and SCRIPTINFO_IUNKNOWN <> 0 then begin
            ppiunkItem := TObjectDispatch.Create(ScriptObject) as IInterface;
            result := S_OK;
            EXIT;
        end;

        if dwReturnMask and SCRIPTINFO_ITYPEINFO <> 0 then begin
            Result := TYPE_E_ELEMENTNOTFOUND;
            EXIT;
        end;
    end;
end;
function TJScript.GetLCID(out plcid: LongWord): HResult;
begin
    plcid := Windows.GetThreadLocale;
    result := S_OK;
end;
function TJScript.OnEnterScript: HResult;
begin
    result := S_OK;
end;
function TJScript.OnLeaveScript: HResult;
begin
    result := S_OK;
end;
function TJScript.OnScriptError(const pscripterror: IActiveScriptError): HResult;
var
    Info: EXCEPINFO;
    scontext: cardinal;
begin
    terminated := true;
    if ActiveX.Succeeded(pscripterror.GetExceptionInfo(Info)) and
        ActiveX.Succeeded(pscripterror.GetSourcePosition(scontext, flastLine, fLastPos)) then  begin
        fLastErr := Info.bstrDescription;
    end;
    Result := SCRIPT_E_REPORTED;
end;
function TJScript.OnScriptTerminate(var pvarResult: OleVariant; var pexcepinfo: EXCEPINFO): HResult;
begin
    result := S_OK;
    terminated := true;
end;
function TJScript.OnStateChange(ssScriptState: tagSCRIPTSTATE): HResult;
begin
    result := S_OK;
    if ssScriptState = SCRIPTSTATE_DISCONNECTED then begin
        terminated := True;
    end;
end;

end.
